m1 = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "metadata": {
            "type": "object",
            "properties": {
                "reference": {"type": "string"},
                "doi": {"type": "string"},
                "name": {"type": "string"},
                "notes": {"type": "string"}
            },
            "required": ["name"]
        },
        "potential": {
            "type": "array",
            "items": {"type": "object",
                      "properties": {
                          "type": {"type": "string"},
                          "parameters": {"type": "object"},
                      },
                      "required": ["type", "parameters"]
                      }
        },
        "cutoff": {
            "type": "array",
            "items": {"type": "object",
                      "properties": {
                          "type": {"type": "string"},
                          "parameters": {"type": "object"},
                      },
                      "required": ["type", "parameters"]
                      }
        }
    },
    "required": ["potential", "cutoff"]
}

m2 = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "metadata": {
            "type": "object",
            "properties": {
                "reference": {"type": "string"},
                "doi": {"type": "string"},
                "name": {"type": "string"},
                "notes": {"type": "string"}
            },
            "required": ["name"]
        },
        "potential": {
            "type": "array",
            "items": {"type": "object",
                      "properties": {
                          "type": {"type": "string"},
                          "parameters": {"type": "object"},
                          "cutoff": {"type": "object"}
                      },
                      "required": ["type", "parameters", "cutoff"]
                      }
        },
    },
    "required": ["potential"]
}



import os
import shutil
import hashlib
import warnings
import numpy
from .schema import schema_version


def pprint(rows, include=None, sort_by=None, max_rows=20, max_len=140, file=None):
    """Pretty print `rows` (a list of dicts)"""

    def _tabular(data, max_len=max_len):
        """General function to format `data` list in tabular table"""
        # Predict formatting
        lens = [0 for _ in range(len(data[0]))]
        for entry in data:
            for i, value in enumerate(entry):
                lens[i] = max(lens[i], len(str(value)))
        fmts = [f'{{:{lens[i]}s}}' for i in range(len(lens))]
        fmt = ' '.join(fmts)

        # Store list of lines
        lines = []
        lines.append(fmt.format(*data[0]))
        lines.append('-'*(sum(lens) + len(lens) - 1))
        for entry in data[1:]:
            entry = [str(_) for _ in entry]
            lines.append(fmt.format(*entry))
            if len(lines) > max_rows and max_rows > 0:
                lines.append(f'... {len(data) - max_rows} entries not shown')
                break

        # Limit columns
        if sum(lens) > max_len:
            for i, line in enumerate(lines):
                if i < 2:
                    fill = '     '
                else:
                    fill = ' ... '
                lines[i] = line[:max_len//2] + fill + line[sum(lens) - max_len//2:]
        return lines

    # Format and sort the data
    if include is None:
        columns = set([e for e in rows[0] if not e.startswith('_')])
        for entry in rows:
            new_columns = set([e for e in entry if not e.startswith('_')])
            columns = set.union(columns, new_columns)
        columns = sorted(columns)
    else:
        columns = include

    if sort_by is not None:
        if not (isinstance(sort_by, list) or isinstance(sort_by, tuple)):
            sort_by = [sort_by]
        rows = sorted(rows[1:], key=lambda x: [x[_] for _ in sort_by])

    # Tabularize lines and join them
    rows = [columns] + [[str(entry.get(key)) for key in columns] for entry in rows]
    lines = _tabular(rows)
    print('\n'.join(lines), file=file)

def _wget(url, output_dir):
    """Like wget on the command line"""
    try:
        from urllib.request import urlopen  # Python 3
    except ImportError:
        from urllib2 import urlopen  # Python 2

    basename = os.path.basename(url)
    output_file = os.path.join(output_dir, basename)
    response = urlopen(url)
    length = 16*1024
    with open(output_file, 'wb') as fh:
        shutil.copyfileobj(response, fh, length)

class _objdict(dict):

    """Boots a dict with object-like attribute accessor"""

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]


def tabulate(potential, npoints=10000, rmax=-1.0, rmin=0.0, overshoot=2, **kwargs):
    """
    Tabulate the potential from 0 to `rmax`.

    The potential cutoff is only used to determine `rmax` if this
    is not given. The full potential is tabulated, it is up to the
    calling code to truncate it. We slightly overshoot the
    tabulation, to avoid boundary effects at the cutoff or at
    discontinuities.
    """
    if hasattr(potential, '_params'):
        rmax = potential._params['rcut']
    assert rmax > 0, 'provide rmax'
    rsq = numpy.ndarray(npoints)
    u0 = numpy.ndarray(npoints)
    u1 = numpy.ndarray(npoints)
    u2 = numpy.ndarray(npoints)
    # We overshoot 2 points beyond rmax (cutoff) to avoid
    # smoothing discontinuous potentials.
    # This is necessary also for the Allen Tildesley lookup table,
    # which for any distance within the cutoff will look up two
    # points forward in the table.
    # Note that the cutoff is applied to the function only to smooth it
    # not to cut it.
    drsq = (rmax**2 - rmin**2) / (npoints - overshoot - 1)
    warnings.filterwarnings("ignore")
    for i in range(npoints):
        rsq[i] = rmin**2 + i * drsq
        # u0[i], u1[i], u2[i] = potential(rsq[i]**0.5)
        # u0[i], u1[i], u2[i] = potential(rsq[i])
        # print(rsq[i]**0.5, u0[i], u1[i], u2[i])
        try:
            u0[i], u1[i], u2[i] = potential(rsq[i]**0.5, **kwargs)
        except ZeroDivisionError:
            u0[i], u1[i], u2[i] = float('nan'), float('nan'), float('nan')
    warnings.resetwarnings()

    # For potentials that diverge at zero, we remove the singularity by hand
    import math
    if math.isnan(u0[0]):
        u0[0], u1[0], u2[0] = u0[1], u1[1], u2[0]
    return rsq, u0, u1, u2


# TODO: clean up and restore the functions below, for internal use

# To add a model
# def add_model_json(path):
#     """
#     If `path` is a directory, add all json files in there to the
#     global `database`. If `path` ends with `json`, it will be assumed
#     to be match one or multiple json files (ex. `*.json`).
#     """
#     import json
#     import glob
#     if path.endswith('json'):
#         search_path = glob.glob(path)
#     else:
#         search_path = glob.glob('{}/*.json'.format(path))

#     for _path in search_path:
#         # Read json file
#         with open(_path) as fh:
#             try:
#                 model = json.load(fh)
#             except (ValueError, json.decoder.JSONDecodeError):
#                 print('Error reading file {}'.format(_path))
#                 raise

#         # By default, the model name is the file basename (stripped of .json)
#         if 'name' not in model:
#             name = os.path.basename(_path)[:-5]
#             model['name'] = '-'.join([entry.capitalize() for entry in name.split('_')])

#         yield model

# To add a sample
# def update(pretend=False):
#     for sample in samples():
#         model, path = sample["model"], sample["path"]
#         sample = {k: sample[k] for k in sample if k not in ["model", "path"]}
#         if path.startswith('http'):
#             store(path, model, pretend=pretend, **sample)
#         else:
#             store(os.path.join(_storage, path), model, pretend=pretend, **sample)


# def _store(path, model, version=0, format=None, notes="", state="", pretend=False, **kwargs):
#     """
#     Keyword arguments can be added to describe the thermodynamic state
#     of the system.
#     """
#     import tempfile
#     _storage_path = '{model}-{variant}-{md5_hash}'

#     # Set paths: storage_path is what goes in the db
#     if path.startswith('http'):
#         tmpdir = tempfile.mkdtemp()
#         basename = os.path.basename(path)
#         _wget(path, tmpdir)
#         local_path = os.path.join(tmpdir, basename)
#     else:
#         local_path = path

#     # We now have a local copy in local_path, get the md5 hash
#     with open(local_path, "rb") as fh:
#         data = fh.read()
#         md5_hash = hashlib.md5(data).hexdigest()
#         extension = os.path.splitext(local_path)[-1]

#     _locals = {}
#     _locals.update(kwargs)
#     _locals['md5_hash'] = md5_hash
#     _locals['model'] = model
#     _locals['version'] = version
#     storage_path = _storage_path.format(**_locals)

#     # Add the entry and its metadata
#     entry = {}
#     entry.update(**kwargs)
#     entry.update({
# 	"model": model,
#         "version": version,
#         "md5_hash": md5_hash,
#         "path": path,
# 	"format": format,
#     })
#     if not path.startswith('http'):
#         entry["path"] = storage_path + extension

#     if pretend:
#         print(os.path.join(_storage, storage_path))
#         return

#     from atooms.core.utils import mkdir
#     mkdir(os.path.dirname(os.path.join(_storage, storage_path)))

#     # Store json file
#     with open(os.path.join(_storage, storage_path + ".json"), 'w') as f:
#         json.dump(entry, f, indent=4)

#     # Store the json file in the storage/ (or samples/).
#     # I would like to use a descriptive path name, but it could be
#     # that the kwargs are not enough to distinguish them.
#     # So I use the md5 hash itself, but prefixed with model name and
#     # version, to have ar least a feeling of what is it.

#     # TODO: issue when having two entries that point to the same file,
#     # but one is a local file reference (within the repo) and the
#     # other one is via the url (on framagit).
#     # But in principle, one could have multiple ways to get to a sample (for redundancy)
#     # If they are all mapped to the same .json file, it does not work.
#     # 1) add a warning that we are overwriting
#     # 2) perhaps add some reference in the path to the storage (file or http)
#     # Actually, there could be multiple instances over the network,
#     # one would need to hash the url!
#     if not path.startswith('http'):
#         with open(os.path.join(_storage, entry["path"]), "wb") as fh:
#             fh.write(data)



class UnsupportedPotential(Exception):
    pass


_map = {
    'lennard_jones': 'lj',
    # 'yukawa': 'yukawa',
    'gaussian': 'gauss',
    'cut_shift': 'cut',
    'cut': 'cut',
}

def _fix_gaussian(parameters):
    """Fix sign convention for Gaussian potential"""
    import copy
    import numpy
    parameters = copy.deepcopy(parameters)
    parameters['epsilon'] = - numpy.array(parameters['epsilon'])
    return parameters

# TODO: tabulate missing potentials


_hack = {
    'gaussian': _fix_gaussian
}

_warn = ['cubic_spline', 'quadratic_cut_shift']

def export(model, tabulate=False):
    """Export model as a string with LAMMPS commands"""
    from . import models
    if not hasattr(model, 'get'):
        model = models.get(model, schema_version=1)
    if not tabulate:
        try:
            return _potential_v1(model)
        except UnsupportedPotential:
            pass
    return _tabulate(model)

    # if models.schema_version(model) == 1:
    #     return _potential_v1(model)
    # else:
    #     raise ValueError('unsupported schema for model')

def _tabulate_lammps(potential, npoints=10000, rmin=0.0, rmax=0.0,
                     metadata='', precision=14, **kwargs):
    """Tabulate a potential."""
    from .helpers import tabulate
    rsq, u0, u1, u2 = tabulate(potential, npoints=npoints, rmin=rmin,
                               rmax=rmax, **kwargs)
    r = rsq**0.5
    # The potential returns -u'/r so u*r will give us the force
    u1 *= r
    txt = f"""

POTENTIAL
N {len(rsq)}

"""
    i = 1
    for x, y, z in zip(r, u0, u1):
        txt += f'{i} {x} {y} {z}\n'
        i += 1
    return txt


def _tabulate(model):
    # TODO: refactor
    assert len(model.get('potential')) == 1
    assert len(model.get('cutoff')) == 1
    phi = model.get('potential')[0].get('type')
    cut = model.get('cutoff')[0].get('type')
    phi_params = model.get('potential')[0].get('parameters')
    cut_params = model.get('cutoff')[0].get('parameters')
    from . import potentials, cutoffs
    nsp = _guess_species(phi_params)
    import os
    import tempfile
    tmpdir = tempfile.mkdtemp()
    base = os.path.join(tmpdir, 'phi')
    # TODO: parametrize npoints
    cmd = ['pair_style table spline 10000']
    pair_coeff = []
    for i in range(nsp):
        for j in range(i, nsp):
            ps = {}
            for key in sorted(phi_params):
                try:
                    ps[key] = phi_params[key][i][j]
                except TypeError:
                    ps[key] = phi_params[key]
            # ps = {key: phi_params[key][i][j] for key in sorted(phi_params)}
            cs = {key: cut_params[key][i][j] for key in sorted(cut_params)}
            # We must redecorate the function every time
            func = potentials[phi]
            cutoff = cutoffs[cut]
            _cutoff = cutoff(func, params=ps, cutoff_params=cs)
            func = _cutoff(func)
            out = _tabulate_lammps(func, rmin=0.1, npoints=10000, overshoot=0, **ps)
            pot_file = f'{base}.{i+1}-{j+1}'
            with open(pot_file, 'w') as fh:
                fh.write(out)
            cmd.append(f'pair_coeff {i+1} {j+1} {pot_file} POTENTIAL')
    return '\n'.join(cmd).strip()

def _guess_species(phi_params):
    nsp = 0
    for key in phi_params:
        try:
            nsp = len(phi_params[key])
            return nsp
        except TypeError:
            continue
    raise ValueError('could not find number of species')

def _potential_v1(model):
    # At this stage we expect a model dictionary
    # As of now, we only support one potential with lammps
    assert len(model.get('potential')) == 1
    assert len(model.get('cutoff')) == 1
    phi = model.get('potential')[0].get('type')
    cut = model.get('cutoff')[0].get('type')
    phi_params = model.get('potential')[0].get('parameters')
    cut_params = model.get('cutoff')[0].get('parameters')
    if phi in _hack:
        phi_params = _hack[phi](phi_params)

    nsp = _guess_species(phi_params)
    try:
        _phi = _map[phi]
        _cut = _map[cut]
    except KeyError:
        raise UnsupportedPotential
    # We must set a dummy cutoff here, but it will be overwritten later
    pair_style = f'pair_style {_phi}/{_cut} 1.0'
    # Some styles do not accept the cut option
    # and only have a global cutoff
    pair_coeff = []
    for i in range(nsp):
        for j in range(i, nsp):
            ps = []
            for entry in sorted(phi_params):
                ps.append(phi_params[entry][i][j])
            ps.append(cut_params['rcut'][i][j])
            ps = ' '.join([str(_) for _ in ps])
            pair_coeff.append(f'pair_coeff {i+1} {j+1} {ps}')
    pair_coeff = '\n'.join(pair_coeff)
    pair_modify = ''
    if 'shift' in cut:
        pair_modify = 'pair_modify shift yes'

    cmd = '\n'.join([pair_style, pair_coeff, pair_modify]).strip()
    return cmd



#!/usr/bin/env python

import os
import unittest
from berni import models, samples, schema

root = os.path.dirname(__file__)

class Test(unittest.TestCase):

    def setUp(self):
        pass

    def _compare_data(self, ref, copy):
        with open(ref) as fh:
            data = fh.read()
        with open(copy) as fh:
            data_copy = fh.read()
        self.assertEqual(data, data_copy)

    def test_grammar(self):
        models.all()
        models.get("lennard_jones")
        samples.all()
        samples.get("lennard_jones-5cc3b80bc415fa5c262e83410ca65779.xyz")

    def test(self):
        sample = samples.get("lennard_jones-5cc3b80bc415fa5c262e83410ca65779.xyz")
        raw = root + '/../berni/samples/lennard_jones-5cc3b80bc415fa5c262e83410ca65779.xyz'
        self._compare_data(raw, sample)

    def test_convert(self):
        m = {
            "metadata": {'name': 'lj'},
            "potential": [
                {
                    "type": "lennard_jones",
                    "parameters": {"sigma": [[1.0, 0.8], [0.8, 0.88]],
                                   "epsilon": [[1.0, 1.5], [1.5, 0.5]]}
                }
            ],
            "cutoff": [
                {
                    "type": "cut_shift",
                    "parameters": {"rcut": [[2.5, 2.0], [2.0, 2.2]]}
                }
            ]
        }
        n = {'potential': [{'cutoff': {'parameters': {'1-1': {'rcut': 2.5},
                                                      '1-2': {'rcut': 2.0},
                                                      '2-2': {'rcut': 2.2}},
                                       'type': 'cut_shift'},
                            'parameters': {'1-1': {'epsilon': 1.0, 'sigma': 1.0},
                                           '1-2': {'epsilon': 1.5, 'sigma': 0.8},
                                           '2-2': {'epsilon': 0.5, 'sigma': 0.88}},
                            'type': 'lennard_jones'}],
             "metadata": {'name': 'lj'},
             }
        self.assertEqual(m, schema._convert(m, 1))
        self.assertEqual(n, schema._convert(m, 2))

    def test_schemas(self):
        from jsonschema import validate
        from berni import models

        for version in [1, ]:
            models.default_table_name = version
            for model in models:
                # Moving the payload to _model means that we cannot
                # use model right away, we must get it
                validate(instance=models.get(model['name']),
                         schema=schema.schemas[version])

    def test_storage(self):
        import glob
        from berni import Query
        query = Query()
        sample = samples.search((query.model == 'lennard_jones') &
                                (query.density == 1.0))[0]
        name = sample['name']
        self.assertTrue(len(samples.get(name)) > 0)

    def test_pprint(self):
        with open('/dev/null', 'w') as null:
            models.pprint(file=null)
            models.pprint(include=['name', 'doi'], file=null)

    def test_potentials(self):
        from berni import potentials
        for f in potentials.values():
            f(0.9)

    def test_tabulate(self):
        from berni.helpers import tabulate
        from berni import potentials, cutoffs

        func = potentials['lennard_jones']
        # print(out)
        # print(func(1.0))
        # cutoff = cutoffs['shift'](func, params={}, cutoff_params={'rcut': [[2.5]]})
        cutoff = cutoffs['shift'](func, params={}, cutoff_params={'rcut': 2.5})
        cut_func = cutoff(func)
        self.assertAlmostEqual(cut_func(1.0)[0], func(1.0)[0] - func(2.5)[0])
        # print(dir(cut_func))
        # print(func(1.0))
        # print(func(1.0)[0])
        # print(cut_func(1.0)[0])
        # print(func(2.5)[0], cut_func(2.5)[0])
        out = tabulate(cut_func, rmax=cut_func._params['rcut'], npoints=100, overshoot=0)
        # print(out[0][-4]**0.5, out[1][-4])
        # print(out[0][-3]**0.5, out[1][-3])
        # print(out[0][-2]**0.5, out[1][-2])
        self.assertAlmostEqual(out[0][-1]**0.5, 2.5)
        self.assertAlmostEqual(out[1][-1], 0.0)

    def test_tabulate_lammps(self):
        import berni.lammps
        from berni import potentials, cutoffs

        func = potentials['lennard_jones']
        cutoff = cutoffs['shift'](func, params={}, cutoff_params={'rcut': 2.5})
        func = cutoff(func)
        out = berni.lammps._tabulate_lammps(func, rmin=0.0, npoints=1000,
                                            overshoot=0)

        model = berni.model('kob_andersen')
        cmd = berni.lammps._tabulate(model)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import atooms.trajectory as trj
import berni
try:
    import rumd
    import berni.rumd
    from atooms.backends.rumd import RUMD
    SKIP = False
except ImportError:
    SKIP = True

class Test(unittest.TestCase):

    def setUp(self):
        if SKIP:
            self.skipTest('missing RUMD')
        self.root = os.path.dirname(__file__)

    def test_(self):
        for model in ['kob_andersen', 'coslovich_pastore']:
            pot = berni.rumd.potential(model)

    def test_cp(self):
        import berni
        inp = berni.samples.get('coslovich_pastore-488db481cdac35e599922a26129c3e35.xyz')
        trj.TrajectoryXYZ(inp).copy(cls=trj.TrajectoryRUMD, fout='/tmp/1.xyz')
        model = 'coslovich_pastore'
        potentials = berni.rumd.potential(model)
        backend = RUMD('/tmp/1.xyz', potentials=potentials)
        epot_0 = backend.system.potential_energy(per_particle=True)

        model = 'coslovich_pastore'
        model = berni.get(model)
        model['cutoff'][0]['type'] = 'linear_cut_shift'
        model['cutoff'][0]['parameters']['rcut'] = model['cutoff'][0]['parameters']['rspl']
        model['cutoff'][0]['parameters'].pop('rspl')
        with trj.Trajectory(inp) as th:
            s = th[0]
            s.species_layout = 'F'
            s.interaction = berni.f90.Interaction(model)
            epot_1 = s.potential_energy(per_particle=True)
        self.assertLess(abs(epot_0 - epot_1), 1e-6)
        import os
        os.system(f'rm -f {inp}')

    def test_ka(self):
        inp = berni.samples.get('kob_andersen-8f4a9fe755e5c1966c10b50c9a53e6bf.xyz')
        trj.TrajectoryXYZ(inp).copy(cls=trj.TrajectoryRUMD, fout='/tmp/1.xyz')
        model = 'kob_andersen'
        potentials = berni.rumd.potential(model)
        backend = RUMD('/tmp/1.xyz', potentials=potentials)
        epot_0 = backend.system.potential_energy(per_particle=True)
        model = 'kob_andersen'
        with trj.Trajectory(inp) as th:
            s = th[0]
            s.species_layout = 'F'
            s.interaction = berni.f90.Interaction(model)
            epot_1 = s.potential_energy(per_particle=True)
        self.assertLess(abs(epot_0 - epot_1), 1e-6)
        import os
        os.system(f'rm -f {inp}')

    def tearDown(self):
        import os
        os.system('rm -f /tmp/1.xyz')


class TestLAMMPS(unittest.TestCase):

    def _system(self, fname):
        with trj.Trajectory(fname) as th:
            system = th[0]
        system.species_layout = 'F'
        return system

    def _backends(self, system, model):
        import atooms.backends.lammps
        import atooms.backends.f90
        import berni.lammps
        import berni.f90
        # LAMMPS
        cmd = berni.lammps.export(model)
        system.interaction = atooms.backends.lammps.Interaction(cmd)
        epot_lammps = system.potential_energy(per_particle=True)

        # LAMMPS tabulated
        cmd = berni.lammps.export(model, tabulate=True)
        system.interaction = atooms.backends.lammps.Interaction(cmd)
        epot_lammps_tab = system.potential_energy(per_particle=True)
        # Atooms
        system.interaction = atooms.backends.f90.Interaction(model)
        epot_f90 = system.potential_energy(per_particle=True)

        self.assertAlmostEqual(epot_lammps, epot_f90)
        self.assertAlmostEqual(epot_lammps, epot_lammps_tab)
        # print(epot_lammps, epot_lammps_tab)

    def _supported(self, system, model, backend):
        import atooms.backends.lammps
        import atooms.backends.f90
        import berni.lammps
        import berni.f90
        if backend == 'lammps':
            cmd = berni.lammps.export(model)
            system.interaction = atooms.backends.lammps.Interaction(cmd)
            epot_lammps = system.potential_energy(per_particle=True)
        if backend == 'f90':
            system.interaction = atooms.backends.f90.Interaction(model)
            epot_f90 = system.potential_energy(per_particle=True)

    def test_lj(self):
        model = berni.get("lennard_jones")
        fname = berni.samples.get("lennard_jones-13ce47602b259f7802e89e23ffd57f19.xyz")
        system = self._system(fname)
        self._backends(system, model)

    def test_ka(self):
        model = berni.models.get('kob_andersen')
        fname = berni.samples.get('kob_andersen-8f4a9fe755e5c1966c10b50c9a53e6bf.xyz')
        system = self._system(fname)
        self._backends(system, model)

    def _backend_support(self, backend):
        """Test all potentials"""
        import inspect
        fname = berni.samples.get("lennard_jones-13ce47602b259f7802e89e23ffd57f19.xyz")
        system = self._system(fname)
        ignore = {
            'f90': ['fene'],
            'lammps': ['fene']
        }
        unsupported, supported = [], []
        for phi, func in berni.potentials.items():
            if phi in ignore['f90']:
                continue

            args = func.__code__.co_varnames[1:]
            args = inspect.getfullargspec(func)[0]
            defaults = inspect.getfullargspec(func)[3]
            params = {arg: [[default]] for arg, default in zip(args[1:], defaults)}

            model = {
                'potential': [{
                    'type': phi,
                    'parameters': params
                }
                ],
                'cutoff': [{
                    'type': 'cut_shift',
                    'parameters': {'rcut': [[2.5]]}
                }
                ]
            }
            # if phi != 'yukawa': continue
            # print(phi)
            # print(model)
            # self._supported(system, model, backend)
            try:
                self._supported(system, model, backend)
                supported.append(phi)
            except:
                unsupported.append(phi)
        return supported, unsupported

    def test_support_f90(self):
        # print('{} {}'.format(*zip(['YES', 'NO'], self._backend_support('f90'))))
        self._backend_support('f90')

    def test_support_lammps(self):
        # print('{} {}'.format(*zip(['YES', 'NO'], self._backend_support('lammps'))))
        self._backend_support('lammps')


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .helpers import _objdict

__all__ = ['shift', 'cut', 'linear', 'quadratic',
           'quadratic_cut_shift', 'linear_cut_shift', 'cut_shift']

class shift:

    # def __init__(self, func, params, cutoff_params):
    #     cutoff_params['rcut'] = np.array(cutoff_params['rcut'])
    #     nsp = cutoff_params['rcut'].shape[0]
    #     cutoff_params['shift'] = np.ndarray((nsp, nsp))
    #     for isp in range(nsp):
    #         for jsp in range(nsp):
    #             cutoff_params['shift'][isp, jsp] = func(cutoff_params['rcut'][isp, jsp], isp, jsp, **params)[0]
    #     # Somehow jax only accepts a reference to a dict
    #     self._params = cutoff_params
    #     self._func = func

    def __init__(self, func, params, cutoff_params):
        cutoff_params['rcut'] = np.array(cutoff_params['rcut'])
        cutoff_params['shift'] = func(cutoff_params['rcut'], **params)[0]
        self._params = cutoff_params
        self._func = func

    def __call__(self, func):
        def wrapper(r, *args, **kwargs):
            return self._func(r, *args, **kwargs) - self._params['shift']
        wrapper._params = self._params
        return wrapper

class cut:
    pass

class linear:
    pass

class quadratic:
    pass

class cubic_spline:
    pass


cut_shift = shift
linear_cut_shift = linear
quadratic_cut_shift = quadratic



try:
    import rumd

except ImportError:
    # Dummy classes for testing
    class _Potential:
        def __init__(self, cutoff_method, **kwargs):
            self.cutoff_method = cutoff_method
            self.__dict__.update(kwargs)
            self.params = {}

        def SetParams(self, i, j, **params):
            self.params[(i, j)] = params

    class _Cutoff:
        def __init__(self, cutoff):
            self.cutoff = cutoff

    class rumd:
        ShiftedForce = _Cutoff
        ShiftedPotential = _Cutoff
        Pot_LJ_12_6 = _Potential
        Pot_IPL_12 = _Potential
        Pot_IPL_n = _Potential
        Pot_Gauss = _Potential


_extra = {'inverse_power': {'n': 'exponent'}}

# TODO: better cut_shift_linear (ad alias)
_map = {
    'cut_shift': rumd.ShiftedPotential,
    'cubic_spline': rumd.ShiftedForce,  # raise WARNING
    'linear_cut_shift': rumd.ShiftedForce,
    'quadratic_cut_shift': rumd.ShiftedForce,  # raise WARNING
    'gaussian': rumd.Pot_Gauss,
    'lennard_jones': rumd.Pot_LJ_12_6,
    # 'inverse_power': rumd.Pot_IPL_12,  # only if exponent 12
    'inverse_power': rumd.Pot_IPL_n
}

_warn = ['cubic_spline', 'quadratic_cut_shift']

# def potential(model):
def export(model):
    """Export model as an array of RUMD potentials"""
    from . import models
    if not hasattr(model, 'get'):
        # This may be a string, so we look for the model in the
        # models database and replace the string with the dictionary
        model = models.get(model, schema_version=2)
    if models.schema_version(model) == 2:
        return _potential_v2(model)
    else:
        raise ValueError('unsupported schema for model')

def _potential_v1(model):
    # At this stage we expect a model dictionary
    assert len(model.get('potential')) == 1
    assert len(model.get('cutoff')) == 1

    n_potentials = len(model.get('potential'))
    potentials = []
    for i in range(n_potentials):
        potential = model.get('potential')[i].get('type')
        potential_parameters = model.get('potential')[i].get('parameters')
        cutoff = model.get('cutoff')[i].get('type')
        cutoff_parameters = model.get('cutoff')[i].get('parameters')

        if potential not in _map:
            raise ValueError('not available {}'.format(potential))
        if cutoff not in _map:
            raise ValueError('not available {}'.format(cutoff))
        if cutoff in _warn:
            import warnings
            warnings.warn("WARNING: incompatible cutoff {} replaced with RUMD cutoff 2 {}".format(cutoff, _map[cutoff]))

        # Some parameters go in the potential constructor
        params = {}
        params['cutoff_method'] = _map[cutoff]
        if potential in _extra:
            for entry in _extra[potential]:
                name = _extra[potential][entry]
                params[entry] = potential_parameters[name]
        pot = _map[potential](**params)

        # To guess the number of species,
        # find the first key that has a list of parameters
        for key in potential_parameters:
            try:
                nsp = len(potential_parameters[key])
                break
            except TypeError:
                continue

        # Loop over species pairs
        for i in range(nsp):
            for j in range(nsp):
                params = {}
                for entry in potential_parameters:
                    Entry = entry.capitalize()
                    # Ignore parameters that entered in the constructor
                    if potential in _extra and entry in _extra[potential].values():
                        continue
                    params[Entry] = potential_parameters[entry][i][j]
                # The input cutoff distance must be divided by sigma for most RUMD potentials
                if 'rcut' in cutoff_parameters:
                    params['Rcut'] = cutoff_parameters["rcut"][i][j] / cutoff_parameters["sigma"][i][j]
                else:
                    # cubic spline has rspl
                    params['Rcut'] = cutoff_parameters["rspl"][i][j] / cutoff_parameters["sigma"][i][j]
                pot.SetParams(i, j, **params)
        potentials.append(pot)

    return potentials

def _potential_v2(model):
    n_potentials = len(model.get('potential'))
    potentials = []
    for i in range(n_potentials):
        potential = model.get('potential')[i].get('type')
        potential_parameters = model.get('potential')[i].get('parameters')
        cutoff = model.get('potential')[i].get('cutoff').get('type')
        cutoff_parameters = model.get('potential')[i].get('cutoff').get('parameters')

        # TODO: refactor
        if potential not in _map:
            raise ValueError('not available {}'.format(potential))
        if cutoff not in _map:
            raise ValueError('not available {}'.format(cutoff))
        if cutoff in _warn:
            import warnings
            warnings.warn("WARNING: incompatible cutoff {} replaced with RUMD cutoff 2 {}".format(cutoff, _map[cutoff]))

        species = []
        first_pair = None
        for pair in potential_parameters:
            if first_pair is None:
                first_pair = pair
            species.append(pair.split('-')[0])
            species.append(pair.split('-')[1])
        species = sorted(set(species))
        nsp = len(species)

        # Some parameters go in the potential constructor
        params = {}
        params['cutoff_method'] = _map[cutoff]
        if potential in _extra:
            for entry in _extra[potential]:
                name = _extra[potential][entry]
                params[entry] = potential_parameters[first_pair][name]
        pot = _map[potential](**params)

        # Loop over species pairs
        for pair in potential_parameters:
            a, b = pair.split('-')
            i, j = species.index(a), species.index(b)
            params = {}
            for entry in potential_parameters[pair]:
                Entry = entry.capitalize()
                # Ignore parameters that entered in the constructor
                if potential in _extra and entry in _extra[potential].values():
                    continue
                params[Entry] = potential_parameters[pair][entry]
            # cubic spline has rspl
            params['Rcut'] = cutoff_parameters[pair]["rcut"] / potential_parameters[pair]["sigma"]
            pot.SetParams(i, j, **params)
        potentials.append(pot)
    return potentials



from . import _schemas

schemas = {1: _schemas.m1, 2: _schemas.m2}

# Model schema helpers

def schema_version(model):
    """
    Return the schema version of the model
    """
    from jsonschema import ValidationError

    # Validate model against either version
    valid = []
    for schema_version in [1, 2]:
        try:
            _validate_model(model, schema_version)
            valid.append(schema_version)
        except ValidationError as e:
            pass
    # Return the schema id
    if len(valid) == 1:
        return valid[0]
    elif len(valid) == 0:
        raise ValidationError(f'invalid model {model}')
    else:
        raise RuntimeError(f'model {model} is valid for multiple schemas, this should not happen')

def _validate_model(model, schema_version=1):
    from jsonschema import validate
    validate(instance=model, schema=schemas[schema_version])

def _convert(model, new_schema_version):
    """
    Convert model to schema `schema_version`. Do nothing is schema version is already the requested one
    """
    if schema_version(model) == new_schema_version:
        return model
    elif schema_version(model) == 1 and new_schema_version == 2:
        return _upgrade_1_to_2(model)
    else:
        raise ValueError('cannot handle this conversion')

def _upgrade_1_to_2(model):
    """Convert from schema version 1 to 2"""
    new_model = {}
    # Optional
    new_model['metadata'] = model['metadata']
    new_model["potential"] = []
    for potential in model["potential"]:
        new_potential = {}
        new_potential["type"] = potential["type"]
        new_potential["parameters"] = {}
        db = {}
        for key in potential["parameters"]:
            db[key] = {}
            nsp = len(potential["parameters"][key])
            for i in range(nsp):
                for j in range(nsp):
                    if j < i:
                        continue
                    pair = f'{i+1}-{j+1}'
                    db[key][pair] = potential["parameters"][key][i][j]
        last_key = key
        for pair in db[last_key].keys():
            new_potential["parameters"][pair] = {key: db[key][pair] for key in db.keys()}
        new_model["potential"].append(new_potential)

    new_cutoffs = []
    for cutoff in model["cutoff"]:
        new_cutoff = {}
        new_cutoff["type"] = cutoff["type"]
        new_cutoff["parameters"] = {}
        db = {}
        for key in cutoff["parameters"]:
            db[key] = {}
            nsp = len(cutoff["parameters"][key])
            for i in range(nsp):
                for j in range(nsp):
                    if j < i:
                        continue
                    pair = f'{i+1}-{j+1}'
                    db[key][pair] = cutoff["parameters"][key][i][j]
        last_key = key
        for pair in db[last_key].keys():
            new_cutoff["parameters"][pair] = {key: db[key][pair] for key in db.keys()}
        new_cutoffs.append(new_cutoff)
    for i, new_cutoff in enumerate(new_cutoffs):
        new_model["potential"][i]["cutoff"] = new_cutoff
    return new_model



# This file is part of atooms
# Copyright 2010-2017, Daniele Coslovich

"""Pair potential classes and factory."""

import warnings
import numpy


def tabulate(potential, npoints=10000, rmax=-1.0, rmin=0.0, overshoot=2, **kwargs):
    """
    Tabulate the potential from 0 to `rmax`.

    The potential cutoff is only used to determine `rmax` if this
    is not given. The full potential is tabulated, it is up to the
    calling code to truncate it. We slightly overshoot the
    tabulation, to avoid boundary effects at the cutoff or at
    discontinuities.
    """
    if hasattr(potential, '_params'):
        rmax = potential._params['rcut']
    assert rmax > 0, 'provide rmax'
    rsq = numpy.ndarray(npoints)
    u0 = numpy.ndarray(npoints)
    u1 = numpy.ndarray(npoints)
    u2 = numpy.ndarray(npoints)
    # We overshoot 2 points beyond rmax (cutoff) to avoid
    # smoothing discontinuous potentials.
    # This is necessary also for the Allen Tildesley lookup table,
    # which for any distance within the cutoff will look up two
    # points forward in the table.
    # Note that the cutoff is applied to the function only to smooth it
    # not to cut it.
    drsq = (rmax**2 - rmin**2) / (npoints - overshoot - 1)
    warnings.filterwarnings("ignore")
    for i in range(npoints):
        rsq[i] = rmin**2 + i * drsq
        # u0[i], u1[i], u2[i] = potential(rsq[i]**0.5)
        # u0[i], u1[i], u2[i] = potential(rsq[i])
        # print(rsq[i]**0.5, u0[i], u1[i], u2[i])
        try:
            u0[i], u1[i], u2[i] = potential(rsq[i]**0.5, **kwargs)
        except ZeroDivisionError:
            u0[i], u1[i], u2[i] = float('nan'), float('nan'), float('nan')
    warnings.resetwarnings()

    # For potentials that diverge at zero, we remove the singularity by hand
    import math
    if math.isnan(u0[0]):
        u0[0], u1[0], u2[0] = u0[1], u1[1], u2[0]
    return rsq, u0, u1, u2


# class PairPotential:

#     def _adjust(self):
#         """Adjust the cutoff to the potential."""
#         self._adjusted = True
#         if self.cutoff is not None and self.cutoff.radius > 0:
#             u = self.func(self.cutoff.radius**2, **self.params)
#             self.cutoff.tailor(self.cutoff.radius**2, u)

#     def compute(self, rsquare):
#         """Compute the potential and its derivatives."""
#         if not self._can_compute:
#             raise ValueError('cannot compute unknown potential %s' % self.func)

#         if not self._adjusted:
#             self._adjust()
#         # Compute the potential and smooth it via the cutoff
#         u = self.func(rsquare, **self.params)
#         if self.cutoff is not None:
#             u = self.cutoff.smooth(rsquare, u)
#         # if rsquare < self.hard_core**2:
#         #     u0, u1 = float("inf"), float("inf")
#         return u
